from __future__ import annotations

import abc
import inspect
from collections import defaultdict
from dataclasses import dataclass
from itertools import chain
from typing import TYPE_CHECKING, Callable, Dict, Generator, Optional, Sequence, Tuple, Type, TypeVar, Union, cast

from docstring_parser import parse as parse_doc
from docstring_parser.parser.common import DocstringParam
from schematics import Model
from schematics.exceptions import DataError

from apistrap.decorators import (
    AcceptsDecorator,
    AcceptsFileDecorator,
    AcceptsQueryStringDecorator,
    IgnoreParamsDecorator,
    RespondsWithDecorator,
    SecurityDecorator,
    TagsDecorator,
)
from apistrap.errors import InvalidFieldsError, InvalidResponseError, UnexpectedResponseError
from apistrap.examples import ExamplesMixin, model_examples_to_openapi_dict
from apistrap.schemas import ErrorResponse
from apistrap.schematics_converters import schematics_model_to_schema_object
from apistrap.tags import TagData
from apistrap.types import FileResponse
from apistrap.utils import resolve_fw_decl, snake_to_camel

if TYPE_CHECKING:  # pragma: no cover
    from apistrap.extension import Apistrap, SecurityScheme


@dataclass(frozen=True)
class ResponseData:
    description: Optional[str] = None
    mimetype: Optional[str] = None


DecoratorType = TypeVar("DecoratorType")


class OperationWrapper(metaclass=abc.ABCMeta):
    """
    Extracts metadata from a view handler function and uses it to 1) generate an OpenAPI specification and 2) implement
    behavior specified by Apistrap decorators.
    """

    def __init__(self, extension: Apistrap, function: Callable, decorators: Sequence[object]):
        """
        Initialize an operation wrapper

        :param extension: The extension that instantiated the wrapper
        :param function: The view handler function to be processed
        :param decorators: The decorators present on the function
        """
        self._responses: Dict[Type[Model], Dict[int, ResponseData]] = defaultdict(lambda: {})
        self._request_body_class: Optional[Type[Model]] = None
        self._request_body_parameter: Optional[str] = None
        self._request_body_file_type: Optional[str] = None
        self._request_body_content_types: Optional[Sequence[str]] = None
        self._path_parameters: Dict[str, Type] = {}
        self._query_parameters: Dict[str, Type] = {}
        self._security: Dict[str, Sequence[str]] = {}
        self._tags: Sequence[str] = []

        self._wrapped_function: Callable = function
        self._decorators: Sequence[object] = decorators
        self._extension = extension
        self._doc = parse_doc(function.__doc__)
        self._signature = inspect.signature(function)

        self.process_metadata()

    def process_metadata(self):
        """
        Load metadata from the view handler function.
        """

        self._responses = self._get_responses()

        (
            self._request_body_parameter,
            self._request_body_class,
            self._request_body_content_types,
        ) = self._get_request_body_parameter()

        if self._request_body_content_types is None:
            self._request_body_content_types = ["application/json"]

        self._request_body_file_type = self._get_request_body_file_type()
        if self._request_body_parameter is not None and self._request_body_file_type is not None:
            raise TypeError("An endpoint cannot accept both a file and a model")

        self._query_parameters = dict(self._get_query_string_parameters())
        self._path_parameters = dict(self._get_path_parameters())

        self._security = [*self._get_security_requirements()]
        self._tags = [*self._get_tags()]

    ##############
    # Public API #
    ##############

    def get_openapi_spec(self):
        """
        Get an OpenAPI Operation object that describes the underlying endpoint.
        """

        spec = {"operationId": snake_to_camel(self._wrapped_function.__name__), "responses": {}}

        if self._doc.short_description:
            spec["summary"] = self._doc.short_description

        if self._doc.long_description:
            spec["description"] = self._doc.long_description

        if self._tags:
            spec["tags"] = self._tags

        if self._path_parameters or self._query_parameters:
            spec["parameters"] = []

        for name, param_type in self._path_parameters.items():
            if self._is_param_ignored(name):
                continue

            param_spec = {
                "name": name,
                "in": "path",
                "required": True,
                "schema": {"type": self._extension.PARAMETER_TYPE_MAP.get(param_type, "string")},
            }

            param_doc = self._get_param_doc(name)
            if param_doc is not None:
                param_spec["description"] = param_doc.description

            spec["parameters"].append(param_spec)

        for name, param_type in self._query_parameters.items():
            param_refl: inspect.Parameter = self._signature.parameters[name]
            param_spec = {
                "name": name,
                "in": "query",
                "required": param_refl.default == inspect.Parameter.empty,
                "schema": {"type": self._extension.PARAMETER_TYPE_MAP.get(param_type, "string")},
            }

            param_doc = self._get_param_doc(name)
            if param_doc is not None:
                param_spec["description"] = param_doc.description

            spec["parameters"].append(param_spec)

        if self._request_body_parameter:
            mimetypes = self._request_body_content_types

            spec["requestBody"] = {
                "content": {
                    mimetype: {"schema": schematics_model_to_schema_object(self._request_body_class, self._extension)}
                    for mimetype in mimetypes
                },
                "required": True,
            }

            if issubclass(self._request_body_class, ExamplesMixin):
                for mimetype in mimetypes:
                    spec["requestBody"]["content"][mimetype]["examples"] = model_examples_to_openapi_dict(
                        self._request_body_class
                    )

            param_doc = self._get_param_doc(self._request_body_parameter)
            if param_doc is not None and param_doc.description:
                spec["requestBody"]["description"] = param_doc.description

            spec["x-codegen-request-body-name"] = "body"
        elif self._request_body_file_type:
            spec["requestBody"] = {
                "content": {self._request_body_file_type: {"schema": {"type": "string", "format": "binary"}}}
            }

        if self._security:
            spec["security"] = self._security

        for response_class, codes in self._responses.items():
            for code, response_data in codes.items():
                if issubclass(response_class, FileResponse):
                    mime = response_data.mimetype or "application/octet-stream"
                    spec["responses"][str(code)] = {
                        "description": response_data.description or response_class.__name__,
                        "content": {mime: {"schema": {"type": "string", "format": "binary"}}},
                    }
                else:
                    spec["responses"][str(code)] = {
                        "description": response_data.description or response_class.__name__,
                        "content": {
                            "application/json": {
                                "schema": schematics_model_to_schema_object(response_class, self._extension)
                            }
                        },
                    }

                    if issubclass(response_class, ExamplesMixin):
                        # fmt: off
                        spec["responses"][str(code)]["content"]["application/json"]["examples"] = \
                            model_examples_to_openapi_dict(response_class)
                        # fmt: on

        return spec

    @abc.abstractmethod
    def get_decorated_function(self):
        """
        Return the view handler decorated with functionality added by Apistrap.
        """

    ######################################
    # Extension points for child classes #
    ######################################

    @abc.abstractmethod
    def is_raw_response(self, response: object) -> bool:
        """
        Check whether the response is raw (which would mean it doesn't need postprocessing).

        :param response: the response to be checked
        :return: True if the response is raw, False otherwise
        """

    #############################
    # Helpers for child classes #
    #############################

    @property
    def accepts_body(self):
        """
        Does the underlying endpoint accept a body parameter
        """
        return self._request_body_parameter is not None

    def _load_request_body(self, body_primitive) -> Dict[str, Model]:
        """
        Load the request body as an object with a fixed schema from a primitive data object (dict structure).

        :param body_primitive: the request body as a primitive object
        :return: the request body as an object of a request class
        """
        if self._request_body_parameter is None or self._request_body_class is None:
            raise ValueError("The endpoint doesn't accept a request body")

        body = self._request_body_class.__new__(self._request_body_class)

        try:
            body.__init__(body_primitive, validate=True, partial=False, strict=True)
        except DataError as ex:
            raise InvalidFieldsError(ex.errors) from ex

        return {self._request_body_parameter: body}

    def _get_required_scopes(self) -> Generator[Tuple[SecurityScheme, Sequence[str]]]:
        """
        Get a list of scopes required by the endpoint.
        """
        for security_decorator in self._find_decorators(SecurityDecorator):
            if (
                len(self._extension.security_schemes) > 1
                and self._extension.default_security_scheme is None
                and security_decorator.security_scheme is None
            ):
                raise TypeError(
                    "Multiple security schemes are defined and no default is set - cannot use security decorator without an explicit scheme"
                )

            if len(self._extension.security_schemes) == 0:
                raise TypeError("At least one security scheme must be defined in order to use the security decorator")

            scheme = (
                security_decorator.security_scheme
                or self._extension.default_security_scheme
                or self._extension.security_schemes[0]
            )
            yield scheme, security_decorator.scopes

    def _postprocess_response(self, response: Union[Model, Tuple[Model, int]]) -> Tuple[Model, int, Optional[str]]:
        """
        Check response type and code and add the code if necessary.
        
        :param response: response received from a view handler
        :return: a response and status code
        """
        code = None

        if isinstance(response, tuple):
            response, code = response

        if self.is_raw_response(response):
            return response, code or 200, ""

        if type(response) not in self._responses.keys():
            raise UnexpectedResponseError(type(response))

        if code is None:
            if len(self._responses[type(response)]) > 1:
                raise InvalidResponseError({"status_code": ["Missing status code"]})
            code = next(iter(self._responses[type(response)].keys()))

        if code not in self._responses[type(response)].keys():
            raise UnexpectedResponseError(type(response), code)

        if isinstance(response, Model):
            try:
                response.validate()
            except DataError as ex:
                raise InvalidResponseError(ex.errors) from ex

        return response, code, self._responses[type(response)][code].mimetype

    def _get_param_doc(self, param_name: str) -> Optional[DocstringParam]:
        """
        Get parameter documentation from the docblock of the underlying view handler function.

        :param param_name: name of the parameter
        :return: the parameter documentation
        """
        for param in self._doc.params:
            if param.arg_name == param_name:
                return param

        return None

    ###################################
    # Extraction of endpoint metadata #
    ###################################

    def _find_decorators(self, decorator_class: Type[DecoratorType]) -> Generator[DecoratorType, None, None]:
        """
        Look up decorators of the view handler by type.
        """
        for decorator in self._decorators:
            if isinstance(decorator, decorator_class):
                yield decorator

    def _get_responses(self):
        """
        Find all possible response classes and codes for the underyling endpoint.
        """
        result: Dict[Type[Model], Dict[int, ResponseData]] = defaultdict(lambda: {})

        for response_class, code, data in chain(
            self._get_response_from_annotation(),
            self._get_responses_from_decorators(),
            self._get_responses_from_raises(),
        ):
            if code in result[response_class].keys():
                raise TypeError("Multiple responses declared with the same schema and code")

            result[response_class][code] = data

        return result

    def _get_response_from_annotation(self) -> Generator[Tuple[Type[Model], int, ResponseData], None, None]:
        """
        Get the response class specified by the return value annotation of the view handler.
        """
        annotation = self._signature.return_annotation

        if annotation is inspect.Signature.empty:
            return

        annotation = resolve_fw_decl(self._wrapped_function, annotation)

        if not issubclass(annotation, Model):
            raise TypeError("Unsupported return type")

        yield annotation, 200, ResponseData(self._doc.returns.description if self._doc.returns else None)

    def _get_responses_from_decorators(self) -> Generator[Tuple[Type[Model], int, ResponseData], None, None]:
        """
        Get the response classes specified by @responds_with decorators.
        """
        for decorator in self._find_decorators(RespondsWithDecorator):
            yield decorator.response_class, decorator.code, ResponseData(decorator.description, decorator.mimetype)

    def _get_responses_from_raises(self) -> Generator[Tuple[Type[Model], int, ResponseData], None, None]:
        """
        Get response classes specified by 'raises' blocks in the docblock of the view handler.
        """
        for item in self._doc.raises:
            exception_type = cast(Type[Exception], resolve_fw_decl(self._wrapped_function, item.type_name))
            code = self._extension.exception_to_http_code(exception_type)

            if code is None:
                continue

            yield ErrorResponse, code, ResponseData(item.description)

    def _get_request_body_parameter(self) -> Union[Tuple[str, Type, Optional[Sequence[str]]], Tuple[None, None, None]]:
        """
        Get the name and type of the parameter used to pass the request body to the view handler and a list of content
        types supported by the handler.
        """
        accepts_decorator = None

        for decorator in self._find_decorators(AcceptsDecorator):
            if accepts_decorator is not None:
                raise TypeError("Multiple @accepts decorators")

            accepts_decorator = decorator

        body_param: Optional[inspect.Parameter] = None
        param: inspect.Parameter
        for param in self._signature.parameters.values():
            annotation = resolve_fw_decl(self._wrapped_function, param.annotation)
            matches_decorator = accepts_decorator and accepts_decorator.request_class == annotation
            is_model = (not accepts_decorator) and issubclass(annotation, Model)

            if matches_decorator or is_model:
                if body_param is not None:
                    if not accepts_decorator:
                        raise TypeError(
                            "Multiple candidates for request body injection. Use @accepts to specify which one should be used."
                        )
                    else:
                        raise TypeError(
                            f"Multiple parameters of type `{accepts_decorator.request_class.__name__}` specified by @accepts"
                        )

                body_param = param

        if body_param is None:
            if accepts_decorator:
                raise TypeError("No parameter for request body injection")

            return None, None, None

        return (
            body_param.name,
            resolve_fw_decl(self._wrapped_function, body_param.annotation),
            accepts_decorator.mimetypes if accepts_decorator else None,
        )

    def _get_request_body_file_type(self) -> Optional[str]:
        """
        If the endpoint accepts files, get the supported types.
        """
        result = None
        for decorator in self._find_decorators(AcceptsFileDecorator):
            if result is not None:
                raise TypeError("An endpoint cannot accept files of multiple types")

            result = decorator.mime_type

        return result

    def _get_query_string_parameters(self) -> Generator[Tuple[str, Type], None, None]:
        """
        Get names and types of parameters accepted by the endpoint via query string.
        """
        for decorator in self._find_decorators(AcceptsQueryStringDecorator):
            for param in decorator.parameter_names:
                if param not in self._signature.parameters:
                    raise TypeError(f"Unknown parameter `{param}`")

                param_refl: inspect.Parameter = self._signature.parameters[param]
                annotation = resolve_fw_decl(self._wrapped_function, param_refl.annotation)

                if annotation not in (str, int):
                    raise TypeError("Only string and integer query parameters are supported")

                yield param, annotation

    def _get_path_parameters(self) -> Generator[Tuple[str, Type], None, None]:
        """
        Get a list of path parameters accepted by the wrapped endpoint.
        """

    def _get_security_requirements(self) -> Generator[Dict[str, Sequence[str]], None, None]:
        """
        Get a security requirement specification from the endpoint.
        """
        for decorator in self._find_decorators(SecurityDecorator):
            scheme = decorator.security_scheme or self._extension.default_security_scheme

            if scheme is None and len(self._extension.security_schemes) == 1:
                scheme = self._extension.security_schemes[0]

            if scheme is None:
                raise TypeError("No security scheme found")

            yield {scheme.name: [*map(str, decorator.scopes)]}

    def _get_tags(self) -> Generator[str, None, None]:
        """
        Get endpoint tags (used in the specification to group similar endpoints).
        """
        for decorator in self._find_decorators(TagsDecorator):
            for tag in decorator.tags:
                if isinstance(tag, TagData):
                    self._extension.add_tag_data(tag)
                yield str(tag)

    def _is_param_ignored(self, param_name: str) -> bool:
        """
        Should a parameter be ignored when generating the specification?
        """
        for decorator in self._find_decorators(IgnoreParamsDecorator):
            for ignored_param in decorator.ignored_params:
                if ignored_param == param_name:
                    return True

        return False



from dataclasses import dataclass
from typing import Optional


@dataclass
class TagData:
    name: str
    description: Optional[str] = None

    def to_dict(self) -> dict:
        result = {"name": self.name}

        if self.description is not None:
            result["description"] = self.description

        return result

    def __str__(self):
        return self.name



from __future__ import annotations

import asyncio
import inspect
import json
import logging
import mimetypes
import re
from functools import wraps
from os import path
from pathlib import Path
from typing import Any, Awaitable, Callable, Coroutine, Dict, Generator, List, Optional, Sequence, Tuple, Type, Union

import jinja2
from aiohttp import StreamReader, web
from aiohttp.web_exceptions import HTTPError
from aiohttp.web_request import BaseRequest, Request
from aiohttp.web_response import Response
from aiohttp.web_urldispatcher import AbstractRoute, DynamicResource, PlainResource
from aiohttp.web_ws import WebSocketResponse

from apistrap.errors import ApiClientError, UnsupportedMediaTypeError
from apistrap.extension import Apistrap, ErrorHandler, SecurityScheme
from apistrap.operation_wrapper import OperationWrapper
from apistrap.schemas import ErrorResponse
from apistrap.types import FileResponse
from apistrap.utils import format_exception, resolve_fw_decl

SecurityEnforcer = Callable[[BaseRequest, Sequence[str]], Union[None, Awaitable[None]]]


class AioHTTPOperationWrapper(OperationWrapper):
    def __init__(
        self, extension: AioHTTPApistrap, function: Callable, decorators: Sequence[object], route: AbstractRoute
    ):
        self.route = route
        super().__init__(extension, function, decorators)
        self._extension = extension

    def process_metadata(self):
        super().process_metadata()
        self._get_aiohttp_request_param_name()

    def _get_aiohttp_request_param_name(self) -> Optional[str]:
        """
        Find the parameter into which the AioHTTP request object should be injected.
        """

        param: inspect.Parameter
        result = None
        for param in self._signature.parameters.values():
            if param.annotation == inspect.Parameter.empty:
                continue

            annotation = resolve_fw_decl(self._wrapped_function, param.annotation)
            if issubclass(annotation, BaseRequest):
                if result is not None:
                    raise TypeError("Multiple candidates for request parameter")
                result = param.name

        if result is not None:
            return result

        if "request" in self._signature.parameters.keys():
            if self._signature.parameters["request"].annotation == inspect.Parameter.empty:
                return "request"

        return None

    async def _enforce_security(self, request):
        error = None

        for security_scheme, required_scopes in self._get_required_scopes():
            try:
                # If any enforcer passes without throwing, the user is authenticated
                enforcer = self._extension.security_enforcers[security_scheme]

                if inspect.iscoroutinefunction(enforcer):
                    await enforcer(request, required_scopes)
                else:
                    enforcer(request, required_scopes)
                return
            except Exception as e:
                error = e
        else:
            if error is not None:
                raise error

    def get_decorated_function(self):
        @wraps(self._wrapped_function)
        async def wrapper(request: Request):
            await self._enforce_security(request)

            kwargs = {}

            if self.accepts_body:
                data = await self._load_request_body_primitive(request)
                kwargs.update(self._load_request_body(data))

            for name, param_type in self._path_parameters.items():
                if name in request.match_info.keys():
                    kwargs[name] = param_type(request.match_info[name])

            for name, param_type in self._query_parameters.items():
                if name in request.rel_url.query.keys():
                    kwargs[name] = param_type(request.query[name])
                elif self._signature.parameters[name].default == inspect.Parameter.empty:
                    raise ApiClientError(f"Missing query parameter `{name}`")

            request_param_name = self._get_aiohttp_request_param_name()
            if request_param_name is not None:
                kwargs[request_param_name] = request

            bound = self._signature.bind(**kwargs)

            response, code, mimetype = self._postprocess_response(
                await self._wrapped_function(*bound.args, **bound.kwargs)
            )

            if self.is_raw_response(response) and not isinstance(response, WebSocketResponse):
                response.set_status(code)
                return response

            if isinstance(response, FileResponse):
                return await self._stream_file_response(request, response, code, mimetype)

            return web.Response(text=json.dumps(response.to_primitive()), content_type="application/json", status=code)

        return wrapper

    async def _load_request_body_primitive(self, request: BaseRequest) -> dict:
        if request.content_type == "application/json":
            try:
                data = await request.json()
            except json.decoder.JSONDecodeError as ex:
                raise ApiClientError("The request body must be a JSON object") from ex

            if isinstance(data, str):
                raise ApiClientError("The request body must be a JSON object")

            return data
        elif request.content_type in ("application/x-www-form-urlencoded", "multipart/form-data"):
            return await request.post()

        raise UnsupportedMediaTypeError()

    async def _stream_file_response(
        self, request: BaseRequest, response: FileResponse, code: int, mimetype: str = None
    ):
        """
        Stream a file response to the client
        """

        # TODO consider implementing add_etags, cache_timeout and conditional
        headers = {}

        if mimetype:
            headers["Content-Type"] = mimetype
        elif response.mimetype:
            headers["Content-Type"] = response.mimetype
        elif response.attachment_filename:
            headers["Content-Type"] = mimetypes.guess_type(response.attachment_filename)[0]

        if response.last_modified is not None:
            headers["Last-Modified"] = str(response.last_modified)

        if response.as_attachment:
            if response.attachment_filename is None:
                raise TypeError("Missing attachment filename")

            headers["Content-Disposition"] = f"attachment,filename={response.attachment_filename}"

        if isinstance(response.filename_or_fp, str) or isinstance(response.filename_or_fp, Path):
            return web.FileResponse(response.filename_or_fp, headers=headers, status=code)
        else:
            # The response contains a file object - stream it to the client
            stream = web.StreamResponse(headers=headers, status=code)

            await stream.prepare(request)
            buffer_size = 16536

            while True:
                if isinstance(response.filename_or_fp, StreamReader):
                    chunk = await response.filename_or_fp.read(buffer_size)
                else:
                    chunk = response.filename_or_fp.read(buffer_size)

                if not chunk:
                    await stream.write_eof()
                    break

                await stream.write(chunk)

            return stream

    def _get_path_parameters(self) -> Generator[Tuple[str, Type], None, None]:
        if not isinstance(self.route.resource, DynamicResource):
            return

        for name in re.findall(r"{([a-zA-Z0-9]+[^}]*)}", self.route.resource.get_info()["formatter"]):
            param_type = str

            if name not in self._signature.parameters.keys():
                yield name, param_type
                continue

            param_refl: inspect.Parameter = self._signature.parameters[name]

            if param_refl.annotation != inspect.Parameter.empty and param_refl.annotation is not None:
                param_type = resolve_fw_decl(self._wrapped_function, param_refl.annotation)

            if param_type not in (str, int):
                raise TypeError(f"Unsupported path parameter type `{param_type.__name__}`")

            yield name, param_type

    def is_raw_response(self, response: object) -> bool:
        return isinstance(response, Response) or isinstance(response, web.StreamResponse)


@web.middleware
class ErrorHandlerMiddleware:
    """
    A configurable handler for exceptions raised when processing HTTP requests.
    """

    def __init__(self, apistrap: "AioHTTPApistrap"):
        """
        :param apistrap: The apistrap extension object
        """
        self._apistrap = apistrap

    def handle_error(self, exception: Exception) -> Tuple[ErrorResponse, int]:
        """
        Find a handler for an exception and invoke it, then return an error response.

        :param exception: the exception to be handled
        :return: an ErrorResponse instance
        """

        response = self._apistrap.exception_to_response(exception)

        if response is None:
            raise ValueError(f"Unexpected exception type `{type(exception).__name__}`") from exception

        return response

    async def __call__(
        self, request: BaseRequest, handler: Callable[[BaseRequest], Coroutine[Any, Any, Response]]
    ) -> web.Response:
        """
        Invoke the middleware.

        :param request: the API request (will be handled without any modifications)
        :param handler: the handler function (will be invoked without any modifications)
        :return: either the response of the handler or an error message response
        """
        try:
            return await handler(request)
        except Exception as ex:
            error_response, code = self.handle_error(ex)

            return web.Response(
                text=json.dumps(error_response.to_primitive()), content_type="application/json", status=code
            )


class AioHTTPApistrap(Apistrap):
    def __init__(self):
        super().__init__()
        self.app: web.Application = None
        self.error_middleware = ErrorHandlerMiddleware(self)
        self.security_enforcers: Dict[SecurityScheme, SecurityEnforcer] = {}
        self._jinja_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(path.join(path.dirname(__file__), "templates"))
        )
        self._default_error_handlers = [
            ErrorHandler(HTTPError, lambda exc_type: exc_type.status_code, self._handle_http_error),
            ErrorHandler(UnsupportedMediaTypeError, 415, self._handle_client_error),
            ErrorHandler(ApiClientError, 400, self._handle_client_error),
            ErrorHandler(Exception, 500, self._handle_server_error),
        ]

    def init_app(self, app: web.Application) -> None:
        """
        Bind the extension to an AioHTTP instance.

        :param app: the AioHTTP instance
        """

        self.app = app
        app.middlewares.append(self.error_middleware)
        app.on_startup.append(self._process_routes)

        if self.spec_url is not None:
            app.router.add_route("get", self.spec_url, self._get_spec)

            if self.ui_url is not None:
                for ui_url in (self.ui_url, self.ui_url + "/"):
                    app.router.add_route("get", ui_url, self._get_ui)

            if self.redoc_url is not None:
                for redoc_url in (self.redoc_url, self.redoc_url + "/"):
                    app.router.add_route("get", redoc_url, self._get_redoc)

    def add_security_scheme(self, scheme: SecurityScheme, enforcer: SecurityEnforcer, *, default: bool = False):
        """
        Add a security scheme to be used by the API.

        :param scheme: a description of the security scheme
        :param enforcer: a function that checks the requirements of the security scheme
        :param default: should this be used as the default security scheme?
        """

        self._add_security_scheme(scheme, default)
        self.security_enforcers[scheme] = enforcer

    def _handle_server_error(self, exception):
        """
        Default handler for server errors (500-599).

        :param exception: the exception to be handled
        :return: an ErrorResponse instance
        """
        logging.exception(exception)
        if asyncio.get_running_loop().get_debug():
            return ErrorResponse(dict(message=str(exception), debug_data=format_exception(exception)))
        else:
            return ErrorResponse(dict(message="Internal server error"))

    def _handle_client_error(self, exception):
        """
        Default handler for client errors (400-499).

        :param exception: the exception to be handled
        :return: an ErrorResponse instance
        """
        if asyncio.get_running_loop().get_debug():
            logging.exception(exception)
            return ErrorResponse(dict(message=str(exception), debug_data=format_exception(exception)))
        else:
            return ErrorResponse(dict(message=str(exception)))

    def _handle_http_error(self, exception: Exception):
        """
        Default handler for http errors (e.g. 404).

        :param exception: the exception to be handled
        :return: an ErrorResponse instance
        """
        if not isinstance(exception, HTTPError):
            raise ValueError()  # pragma: no cover

        logging.exception(exception)
        return ErrorResponse(dict(message=exception.text))

    async def _get_spec(self, request: Request):
        """
        Serves the OpenAPI specification
        """

        return web.Response(text=json.dumps(self.to_openapi_dict()), content_type="application/json", status=200)

    _get_spec.apistrap_ignore = True

    async def _get_ui(self, request: Request):
        """
        Serves Swagger UI
        """

        return web.Response(
            text=self._jinja_env.get_template("apidocs.html").render(apistrap=self),
            content_type="text/html",
            status=200,
        )

    _get_ui.apistrap_ignore = True

    async def _get_redoc(self, request: Request):
        """
        Serves ReDoc
        """

        return web.Response(
            text=self._jinja_env.get_template("redoc.html").render(apistrap=self), content_type="text/html", status=200
        )

    _get_redoc.apistrap_ignore = True

    async def _process_routes(self, *args, **kwargs) -> None:
        """
        Process all non-ignored routes and their parameters
        """

        operations = [
            AioHTTPOperationWrapper(self, route._handler, self._get_decorators(route._handler), route)
            for route in self.app.router.routes()
            if not self._is_route_ignored(route.method, route._handler)
        ]

        self._extract_specs(operations)

        for op in operations:
            op.route._handler = op.get_decorated_function()

    def _extract_specs(self, operations: List[AioHTTPOperationWrapper]) -> None:
        """
        Extract the specification data from the bound AioHTTP app. If the data was already extracted, do not do
        anything.
        """

        for op in operations:
            url = ""

            if isinstance(op.route.resource, PlainResource):
                url = op.route.resource.get_info()["path"]
            elif isinstance(op.route.resource, DynamicResource):
                url = op.route.resource.get_info()["formatter"]

            self.spec.path(url, {op.route.method.lower(): op.get_openapi_spec()})

    def _is_bound(self) -> bool:
        return self.app is not None

    def _get_default_error_handlers(self) -> Sequence[ErrorHandler]:
        return self._default_error_handlers



from __future__ import annotations

from inspect import getmro
from typing import TYPE_CHECKING, Any, Dict, Generator, Optional, Type

from schematics import Model
from schematics.types.base import (
    BaseType,
    BooleanType,
    DecimalType,
    FloatType,
    IntType,
    LongType,
    NumberType,
    StringType,
)
from schematics.types.compound import DictType, ListType, ModelType

from apistrap.utils import snake_to_camel

if TYPE_CHECKING:  # pragma: no cover
    from apistrap.extension import Apistrap


def _get_serialized_name(field: BaseType) -> str:
    """
    Get the name under which the model field is saved in the serialized object

    :param field: the model field
    :return: a field name
    """
    return getattr(field, "serialized_name", None) or field.name


def _model_fields_to_schema_object_properties(model: Type[Model], apistrap: Optional[Apistrap]) -> Dict[str, Any]:
    """
    Convert all fields of a model to OpenAPI 3 SchemaObject Properties objects

    :param model: the model to be converted
    :param apistrap: the extension used for adding reusable schema definitions
    :return: a dictionary with field names as keys and SchemaObjects as values
    """
    properties = {}

    for field in model._fields.values():
        serialized_name = _get_serialized_name(field)
        schema_object = _field_to_schema_object(field, apistrap)

        if schema_object is not None:
            properties[serialized_name] = schema_object

    return properties


def _field_to_schema_object(field: BaseType, apistrap: Optional[Apistrap]) -> Optional[Dict[str, Any]]:
    """
    Convert a field definition to OpenAPI 3 schema.

    :param field: the field to be converted
    :param apistrap: the extension used for adding reusable schema definitions
    :return: a schema
    """

    if isinstance(field, ModelType):
        return _model_field_to_schema_object(field, apistrap)
    elif isinstance(field, ListType):
        if isinstance(field.field, ModelType):
            return _model_array_to_schema_object(field, apistrap)
        elif isinstance(field.field, BaseType):
            return _primitive_array_to_schema_object(field)
    elif isinstance(field, DictType):
        if isinstance(field.field, ModelType):
            return _model_dict_to_schema_object(field, apistrap)
        elif isinstance(field.field, BaseType):
            return _primitive_dict_to_schema_object(field)
    elif isinstance(field, StringType):
        return _string_field_to_schema_object(field, apistrap)
    elif isinstance(field, BaseType):
        return _primitive_field_to_schema_object(field)

    return None


SCHEMATICS_TYPE_TO_JSON_TYPE = {
    NumberType: "number",
    IntType: "integer",
    LongType: "integer",
    FloatType: "number",
    DecimalType: "number",
    BooleanType: "boolean",
}

SCHEMATICS_OPTIONS_TO_JSON_SCHEMA = {
    "max_length": "maxLength",
    "min_length": "minLength",
    "regex": "pattern",
    "min_value": "minimum",
    "max_value": "maximum",
    "min_size": "minItems",
    "max_size": "maxItems",
}


def _get_field_type(field: BaseType):
    for cls in getmro(field.__class__):
        if cls in SCHEMATICS_TYPE_TO_JSON_TYPE.keys():
            return SCHEMATICS_TYPE_TO_JSON_TYPE[cls]

    return "string"


def _extract_model_description(field: BaseType) -> Dict[str, Any]:
    """
    Load title and description from Schematics metadata.

    :param field: field to extract metadata from
    :return a dict with optional title and description keys
    """

    metadata = field.metadata
    data = {}

    if "label" in metadata:
        data["title"] = metadata["label"]

    if "description" in metadata:
        data["description"] = metadata["description"]

    return data


def _primitive_field_to_schema_object(field: BaseType) -> Dict[str, str]:
    schema = {"type": _get_field_type(field)}

    schema.update(_extract_model_description(field))

    for schematics_attr, json_schema_attr in SCHEMATICS_OPTIONS_TO_JSON_SCHEMA.items():
        if hasattr(field, schematics_attr):
            option_value = getattr(field, schematics_attr)
            if option_value is not None:
                schema[json_schema_attr] = option_value

    return schema


def _string_field_to_schema_object(field: StringType, apistrap: Optional[Apistrap] = None) -> Dict[str, str]:
    schema = _primitive_field_to_schema_object(field)

    if field.choices is not None:
        if apistrap is None:
            schema["enum"] = field.choices
        else:
            name = snake_to_camel(field.name)
            name = name[0].upper() + name[1:]

            schema = {"$ref": apistrap.add_schema_definition(name, {"type": "string", "enum": field.choices})}

    return schema


def _get_required_fields(model: Type[Model]) -> Generator[str, None, None]:
    """
    Extract required field names from a Schematics model

    :param model:
    :return: a generator of field names
    """

    for field in model._fields.values():
        if getattr(field, "required", False):
            yield _get_serialized_name(field)


def _model_array_to_schema_object(field: ModelType, apistrap: Optional[Apistrap]) -> Dict[str, Any]:
    """
    Get a SchemaObject for a list of given models

    :param field: the field field to be converted
    :param apistrap: the extension used for adding reusable schema definitions
    :return: a SchemaObject
    """

    model = field.field.model_class

    schema = {
        "type": "array",
        "title": f"List of {model.__name__}",
        "items": schematics_model_to_schema_object(model, apistrap),
    }

    schema.update(_extract_model_description(field))

    return schema


def _model_dict_to_schema_object(field: DictType, apistrap: Optional[Apistrap]) -> Dict[str, Any]:
    """
    Get a SchemaObject for a dictionary of given models

    :param field: the field to be converted
    :param apistrap: the extension used for adding reusable schema definitions
    :return: a SchemaObject
    """

    model = field.field.model_class

    schema = {
        "type": "object",
        "title": f"Dictionary of {model.__name__}",
        "additionalProperties": schematics_model_to_schema_object(model, apistrap),
    }

    schema.update(_extract_model_description(field))

    return schema


def _primitive_array_to_schema_object(field: ListType) -> Dict[str, Any]:
    """
    Get a SchemaObject for a list of primitive types

    :param field: the field that determines the value type
    :return: a SchemaObject
    """

    schema = {
        "type": "array",
        "title": f"List of {field.field.__class__.__name__}",
        "items": _field_to_schema_object(field.field, None),
    }

    schema.update(_extract_model_description(field))

    return schema


def _primitive_dict_to_schema_object(field: DictType) -> Dict[str, Any]:
    """
    Get a SchemaObject for a dictionary of primitive types

    :param field: the field that determines the value type
    :return: a SchemaObject
    """

    schema = {
        "type": "object",
        "title": f"Dictionary of {field.field.__class__.__name__}",
        "additionalProperties": _field_to_schema_object(field.field, None),
    }

    schema.update(_extract_model_description(field))

    return schema


def _model_field_to_schema_object(field: ModelType, apistrap: Optional[Apistrap]) -> Dict[str, Any]:
    """
    Get a SchemaObject for a model field.

    :param field: the field that determines the value type
    :param apistrap: the extension used for adding reusable schema definitions
    :return: a SchemaObject
    """

    schema = schematics_model_to_schema_object(field.model_class, apistrap)
    schema.update(_extract_model_description(field))

    return schema


def schematics_model_to_schema_object(model: Type[Model], apistrap: Optional[Apistrap] = None) -> Dict[str, Any]:
    """
    Convert a Schematics model to an OpenAPI 3 SchemaObject

    :param model: the model to be converted
    :param apistrap: the extension used for adding reusable schema definitions
    :return: a SchemaObject
    """

    schema = {
        "type": "object",
        "title": model.__name__,
        "properties": (_model_fields_to_schema_object_properties(model, apistrap)),
    }

    required = list(_get_required_fields(model))

    if required:
        schema["required"] = required

    if apistrap is None:
        return schema  # There is no way to reuse the schema definition

    name = apistrap.add_schema_definition(model.__name__, schema)

    return {"$ref": name}



import inspect
import traceback
from typing import Any, Callable, Dict, Mapping, Type, Union

from more_itertools import flatten


def format_exception(exception: Exception) -> Mapping[str, Any]:
    """
    Format an exception into a dict containing exception information such as class name, message and traceback.

    Example:

        {
            'exception_type': 'RuntimeError'
            'exception_message': 'Cannot connect to database.',
            'traceback': ['  File "spam.py", line 3, in <module>',
                          '   spam.eggs()\n',
                          ...]
        }
    """

    traceback_summary = traceback.StackSummary.extract(traceback.walk_tb(exception.__traceback__))

    return {
        "exception_type": type(exception).__name__,
        "exception_message": str(exception),
        "traceback": list(flatten([str(s).rstrip("\n").split("\n") for s in traceback_summary.format()])),
    }


def snake_to_camel(value):
    """
    Convert a string from snake_case to camelCase
    """
    result = "".join(x.capitalize() or "_" for x in value.split("_"))
    return result[0].lower() + result[1:]


def get_function_perspective_globals(function: Callable) -> Dict[str, Any]:
    """
    Extract global variables from the stack of a function.

    :param function: the function whose stack should be examined.
    :return: a merged dictionary of globals
    """

    globs = {}
    stack = inspect.stack()

    for frame in stack:
        if frame.filename == inspect.getsourcefile(function):
            globs.update(frame.frame.f_globals)

    module = inspect.getmodule(function)
    if module:
        globs.update(inspect.getmembers(module))

    return globs


def resolve_fw_decl(function: Callable, annotation: Union[None, str, Type]) -> Type:
    """
    Resolve a parameter annotation.

    :param function: the function that owns the annotation
    :param annotation: an annotation string
    :return: the resolved annotation
    """
    if annotation is None:
        raise ValueError("NoneType parameter annotations are not supported")

    if not isinstance(annotation, str):
        return annotation

    globs = get_function_perspective_globals(function)

    return eval(annotation, globs)


def get_type_hints(function: Callable) -> Dict[str, Type]:
    """
    Get a dictionary of resolved type annotations for a function.
    """

    return {name: resolve_fw_decl(function, annotation) for name, annotation in function.__annotations__.items()}



import abc
from abc import ABCMeta
from dataclasses import dataclass
from functools import partial
from itertools import chain
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Type, TypeVar, Union

from apispec import APISpec
from apispec.utils import OpenAPIVersion
from schematics import Model

from apistrap.decorators import (
    AcceptsDecorator,
    AcceptsFileDecorator,
    AcceptsQueryStringDecorator,
    IgnoreDecorator,
    IgnoreParamsDecorator,
    RespondsWithDecorator,
    SecurityDecorator,
    TagsDecorator,
)
from apistrap.errors import ApistrapExtensionError
from apistrap.schemas import ErrorResponse
from apistrap.tags import TagData


class SecurityScheme(metaclass=ABCMeta):
    """
    Description of an authentication method.
    """

    def __init__(self, name: str):
        """
        :param name: Name of the scheme (used as the name in the OpenAPI specification)
        :param enforcer: An object invoked by an extension that takes a list of scopes and raises an error if the user
                         doesn't have them
        """
        self.name = name

    @abc.abstractmethod
    def to_openapi_dict(self):
        """
        Get a JSON-serializable object with an OpenAPI description of the security scheme
        """


class OAuthFlowDefinition:
    """
    A holder for the description of an OpenAPI OAuth flow
    """

    def __init__(
        self, flow_type: str, scopes: Dict[str, str], auth_url: Optional[str] = None, token_url: Optional[str] = None
    ):
        """
        :param flow_type: e.g. password, implicit, ...
        :param scopes: a dict of scope name : human-readable description of available scopes
        :param auth_url: OAuth 2 authentication endpoint URL (if applicable)
        :param token_url: OAuth 2 token endpoint URL (if applicable)
        """
        self.flow_type = flow_type
        self.scopes = scopes
        self.auth_url = auth_url
        self.token_url = token_url

    def to_openapi_dict(self):
        """
        :return: A dictionary to be used in the OpenAPI specification
        """
        result = {"scopes": self.scopes}

        if self.auth_url is not None:
            result["authorizationUrl"] = self.auth_url

        if self.token_url is not None:
            result["tokenUrl"] = self.token_url

        return result


class OAuthSecurity(SecurityScheme):
    """
    A description of an OAuth security scheme with an arbitrary list of OAuth 2 flows
    """

    def __init__(self, name: str, *flows: OAuthFlowDefinition):
        """
        :param flows: A list of OAuth 2 flows allowed by the security scheme
        """
        super().__init__(name)
        self.flows = flows

    def to_openapi_dict(self):
        return {"type": "oauth2", "flows": {flow.flow_type: flow.to_openapi_dict() for flow in self.flows}}


ExceptionType = TypeVar("ExceptionType", bound=Exception)


@dataclass
class ErrorHandler:
    exception_class: Type[Exception]
    http_code: Union[int, Callable[[Type[Exception]], int]]
    handler: Callable[[Exception], ErrorResponse]

    def get_code(self, exception_type: Type[Exception]) -> int:
        if callable(self.http_code):
            return self.http_code(exception_type)

        return self.http_code


class Apistrap(metaclass=ABCMeta):
    """
    An abstract ancestor for extensions that bind Apistrap to a web framework
    """

    PARAMETER_TYPE_MAP = {int: "integer", str: "string"}

    def __init__(self):
        self.spec = APISpec(openapi_version=OpenAPIVersion("3.0.2"), title="API created with Apistrap", version="1.0.0")
        self.description = None
        self.oauth_client_id = None
        self.oauth_client_secret = None
        self.security_schemes: List[SecurityScheme] = []
        self.default_security_scheme: Optional[SecurityScheme] = None
        self._spec_url = "/spec.json"
        self._ui_url = "/apidocs"
        self._redoc_url = None
        self._use_default_error_handlers = True
        self._error_handlers: List[ErrorHandler] = []

    def to_openapi_dict(self):
        """
        :return: a dict representation of the OpenAPI spec that can be directly serialized to JSON or YAML
        """

        result = self.spec.to_dict()

        if self.description:
            result["info"]["description"] = self.description

        return result

    ######################################
    # Extension points for child classes #
    ######################################

    @abc.abstractmethod
    def _is_bound(self) -> bool:
        """
        Check whether the extension is bound to an app.
        """

    @abc.abstractmethod
    def _get_default_error_handlers(self) -> Sequence[ErrorHandler]:
        """
        Get a collection of default error handlers
        """

    ###################
    # Utility methods #
    ###################

    def _ensure_not_bound(self, message: str) -> None:
        """
        Throw an exception if the extension is already bound

        :param message: the message of the exception
        """
        if self._is_bound():
            raise ApistrapExtensionError(message)

    def _is_route_ignored(self, method: str, handler) -> bool:
        """
        Check if a view handler should be ignored.

        :param method: the HTTP method
        :param handler: the handler function
        """

        if method.lower() not in ["get", "post", "put", "delete", "patch"]:
            return True

        for decorator in self._get_decorators(handler):
            if isinstance(decorator, IgnoreDecorator):
                return True

        if getattr(handler, "apistrap_ignore", False):
            return True

        return False

    def _exception_handler(self, exception_class: Type[Exception]) -> Optional[ErrorHandler]:
        if self.use_default_error_handlers:
            handlers = chain(self._error_handlers, self._get_default_error_handlers())
        else:
            handlers = self._error_handlers

        for handler in handlers:
            if issubclass(exception_class, handler.exception_class):
                return handler

        return None

    def exception_to_http_code(self, exception_class: Type[Exception]) -> Optional[int]:
        handler = self._exception_handler(exception_class)

        if handler is None:
            return 500

        return handler.get_code(exception_class)

    def exception_to_response(self, exception: Exception) -> Optional[Tuple[ErrorResponse, int]]:
        handler = self._exception_handler(type(exception))

        if handler is None:
            return None

        return handler.handler(exception), handler.get_code(type(exception))

    @property
    def has_error_handlers(self) -> bool:
        return self.use_default_error_handlers or self._error_handlers

    ########################################
    # Configuration methods and properties #
    ########################################

    @property
    def title(self) -> str:
        """
        A title of the OpenAPI specification (the name of the API)
        """
        return self.spec.title

    @title.setter
    def title(self, title: str):
        self.spec.title = title

    @property
    def spec_url(self) -> Optional[str]:
        """
        The URL where the extension should serve the OpenAPI specification. If it is None, the specification is not
        served at all.
        """
        return self._spec_url

    @spec_url.setter
    def spec_url(self, value):
        self._ensure_not_bound("You cannot configure the spec_url after binding the extension with an app")
        self._spec_url = value

    @property
    def ui_url(self) -> Optional[str]:
        """
        The URL where the extension should serve the Swagger UI. If it is None (or if the specification is not served),
        the UI is not served at all.
        """
        return self._ui_url

    @ui_url.setter
    def ui_url(self, value: Optional[str]):
        self._ensure_not_bound("You cannot change the UI url after binding the extension with an app")
        self._ui_url = value.rstrip("/") if value is not None else None

    @property
    def redoc_url(self) -> Optional[str]:
        """
        The URL where the extension should serve the ReDoc documentation. If it is None (or if the specification is not
        served), the UI is not served at all.
        """
        return self._redoc_url

    @redoc_url.setter
    def redoc_url(self, value: Optional[str]):
        self._ensure_not_bound("You cannot change the ReDoc url after binding the extension with an app")
        self._redoc_url = value.rstrip("/") if value is not None else None

    @property
    def use_default_error_handlers(self) -> bool:
        """
        A flag that indicates if the extension should register its error handlers when binding it with an app
        """
        return self._use_default_error_handlers

    @use_default_error_handlers.setter
    def use_default_error_handlers(self, value: bool):
        self._ensure_not_bound("You cannot change the error handler settings after binding the extension with an app")
        self._use_default_error_handlers = value

    def add_error_handler(
        self,
        exception_class: Type[ExceptionType],
        http_code: Union[int, Callable[[Type[ExceptionType]], int]],
        handler: Callable[[ExceptionType], ErrorResponse],
    ) -> None:
        """
        Add an error handler function.
        """

        self._ensure_not_bound("You cannot add error handlers after binding the extension with an app")
        self._error_handlers.append(ErrorHandler(exception_class, http_code, handler))

    ###################################
    # Component definition management #
    ###################################

    def add_request_definition(self, name: str, schema: dict):
        """
        Add a new request definition to the specification. If a different schema is supplied for an existing definition,
        a ValueError is raised.

        :param name: the name of the definition (without the '#/components/.../' part)
        :param schema: a JsonObject OpenAPI structure
        :return: the full path to the definition in the specification file (can be used directly with $ref)
        """

        components = self.spec.components.to_dict()
        if "schemas" in components and name in components["schemas"]:
            if components["schemas"][name] != schema:
                raise ValueError(f"Conflicting definitions of `{name}`")
        else:
            self.spec.components.schema(name, schema)

        return f"#/components/schemas/{name}"

    def add_response_definition(self, name: str, schema: dict):
        """
        Add a new response definition to the specification. If a different schema is supplied for an existing definition,
        a ValueError is raised.

        :param name: the name of the definition (without the '#/components/.../' part)
        :param schema: a JsonObject OpenAPI structure
        :return: the full path to the definition in the specification file (can be used directly with $ref)
        """

        components = self.spec.components.to_dict()
        if "responses" in components and name in components["responses"]:
            if components["responses"][name] != schema:
                raise ValueError(f"Conflicting definitions of `{name}`")
        else:
            self.spec.components.response(name, schema)

        return f"#/components/responses/{name}"

    def add_schema_definition(self, name: str, schema: dict):
        """
        Add a new schema definition to the specification. If a different schema is supplied for an existing definition,
        a ValueError is raised.

        :param name: the name of the definition (without the '#/components/.../' part)
        :param schema: a JsonObject OpenAPI structure
        :return: the full path to the definition in the specification file (can be used directly with $ref)
        """

        components = self.spec.components.to_dict()
        if "schemas" in components and name in components["schemas"]:
            if components["schemas"][name] != schema:
                raise ValueError(f"Conflicting definitions of `{name}`")
        else:
            self.spec.components.schema(name, schema)

        return f"#/components/schemas/{name}"

    def add_tag_data(self, tag: TagData) -> None:
        """
        Add information about a tag to the specification.
        :param tag: data about the tag
        """

        spec = self.spec.to_dict()
        if "tags" not in spec or tag.name not in map(lambda t: t["name"], spec["tags"]):
            self.spec.tag(tag.to_dict())

    def _add_security_scheme(self, scheme: SecurityScheme, default: bool):
        self.spec.components.security_scheme(scheme.name, scheme.to_openapi_dict())
        self.security_schemes.append(scheme)
        if default:
            if self.default_security_scheme is not None:
                raise ApistrapExtensionError("A default security scheme is already set")

            self.default_security_scheme = scheme

    #######################
    # Decorator utilities #
    #######################

    DECORATORS_ATTR = "__apistrap_decorators"

    def _decorate(self, decorator: object, function: Callable):
        if not hasattr(function, self.DECORATORS_ATTR):
            setattr(function, self.DECORATORS_ATTR, [])

        self._get_decorators(function).append(decorator)

        return function

    def _get_decorators(self, function: Callable):
        return getattr(function, self.DECORATORS_ATTR, [])

    #######################
    # Decorator factories #
    #######################

    def tags(self, *tags: Union[str, TagData]):
        """
        A decorator that adds tags to the OpenAPI specification of the decorated view function.
        """
        return partial(self._decorate, TagsDecorator(tags))

    def ignore(self):
        """
        A decorator that marks an endpoint as ignored so that the extension won't include it in the specification.
        """
        return partial(self._decorate, IgnoreDecorator())

    def ignore_params(self, *ignored_params: str):
        """
        A decorator that tells Apistrap to ignore given parameter names when generating documentation for an operation.
        """
        return partial(self._decorate, IgnoreParamsDecorator(ignored_params))

    def security(self, *scopes: str, scheme: SecurityScheme = None):
        """
        A decorator that enforces user authentication and authorization.
        """
        return partial(self._decorate, SecurityDecorator(scopes, security_scheme=scheme))

    def responds_with(
        self,
        response_class: Type[Model],
        *,
        code: int = 200,
        description: Optional[str] = None,
        mimetype: Optional[str] = None,
    ):
        """
        A decorator that fills in response schemas in the Swagger specification. It also converts Schematics models
        returned by view functions to JSON and validates them.
        """
        return partial(self._decorate, RespondsWithDecorator(response_class, code, description, mimetype))

    def accepts(self, request_class: Type[Model], mimetypes: Optional[Sequence[str]] = None):
        """
        A decorator that validates request bodies against a schema and passes it as an argument to the view function.
        The destination argument must be annotated with the request type.
        """
        return partial(
            self._decorate, AcceptsDecorator(request_class, mimetypes) if mimetypes else AcceptsDecorator(request_class)
        )

    def accepts_file(self, mime_type: str = "application/octet-stream"):
        """
        A decorator used to declare that an endpoint accepts a file as the request body.
        """
        return partial(self._decorate, AcceptsFileDecorator(mime_type))

    def accepts_qs(self, *parameter_names: str):
        """
        A decorator used to declare that an endpoint accepts query string parameters.
        """
        return partial(self._decorate, AcceptsQueryStringDecorator(parameter_names))
